#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "emmathfnt.h"

/* Boolean macros to make logical operations more readable */ 

#define TRUE 1
#define FALSE 0

/* 
* 
* Quoted from: PostScript Language Document Structuring Conventions 
* Specification, Version 3.0
*
* Line Length
* To provide compatibility with a large body of existing application 
* and document manager software, a conforming PostScript language document
* description does not have lines exceeding 255 characters, excluding 
* line-termination characters. The intent is to be able to read lines 
* into a 255-character buffer without overflow (Pascal strings are a 
* common example of this sort of buffer).
*
* This macro is used in memory allocation and file retrieval functions.
*
*/

#define PS_BUFSIZE 257

/* We assume that ASCII versions of the Type 1 fonts are to be used. */

#define FONT_EXT ".pfa"

/*
*
* These macros make the code that packs the master font list structure
* a little easier to read. 
*
*/

#define N_FONT_FILS N_FONT_FAMS*N_FONT_FACS
#define N_FONT_FAMS 5 	/* Math1 - Math 5 */
#define N_FONT_FACS 4	/* regular, mono, bold, mono-bold */
#define FONT_FAMILY "Math"
#define FONT_FACE_LIST {"", "-Bold", "Mono", "Mono-Bold"}

/* Sets aside one character for storing the font family number. */

#define FAM_NUMBER 1

/* 
*
* Used in calls to modify the flag denoting that the font is 
* to be used in the file. 
*
*/

#define FONT_IN_USE 1

/* Macros for different resource comment types. */

#define DOC_NEED_RES "%%DocumentNeededResources"
#define DOC_NEED_FNT "%%DocumentNeededFonts"
#define DOC_SUPP_RES "%%DocumentSuppliedResources"
#define DOC_SUPP_FNT "%%DocumentSuppliedFonts"
#define CONTINUATION "%%+"

#define INCL_RES "%%IncludeResource: font " FONT_FAMILY
#define INCL_FNT "%%IncludeFont"

#define BEG_RES "%%BeginResource: font "
#define BEG_FNT "%%BeginFont: "
#define END_RES "%%EndResource\n"
#define END_FNT "%%EndFont\n"

/*
*
* Codes used to identify different types of comments.  Reduces the
* number of calls to string comparison functions. 
*
*/

#define CODE_NO_COMMENT		0
#define CODE_NEED_RES		1
#define CODE_NEED_FNT		2
#define CODE_SUPP_RES		3
#define CODE_SUPP_FNT		4
#define CODE_CONTINUATION	5
#define CODE_OTHER			6

/* Default directory locations for the Math fonts. */

#if unix || __unix || __unix__ ||_AIX
#  define MATH_DIR "/usr/local/mathematica/"
#  define DEF_FONT_DIR MATH_DIR "SystemFiles/Fonts/Type1/"
#elif defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#  define MATH_DIR "C:\\Program Files\\Wolfram Research\\Mathematica\\4.0\\"
#  define DEF_FONT_DIR MATH_DIR "SystemFiles\\Fonts\\Type1\\"
#elif macintosh || Macintosh || THINK_C || defined(_MAC) || defined(__MRC__)
#  define MATH_DIR "Macintosh HD:Mathematica 4.0 Files:"
#  define DEF_FONT_DIR MATH_DIR "SystemFiles:Fonts:Type1:"
#endif

/*
*
* Whether a Math font is to be embedded in an EPS file is logged
* in the fontdata struct.  It contains a character array that stores
* the PostScript name of the font and a boolean flag that declares
* whether the font is in use.
*
*/

struct fontdata {
	char* desc;
	char inuse;
};

/* Exported function prototype, defined in emmathfnt.h */

/* int emmathfnt(FILE* stmin, FILE* stmout, char* fontdir); */

/* Private functions */

void packfontlist(struct fontdata*);
struct fontdata* findfont(struct fontdata*, char*);
void setfontflag(struct fontdata*, char);
char isfontflagset(struct fontdata*);
char* fonttofilename(struct fontdata*, char*); 
char* tokenget(char*, int);
int rewriteheader(FILE* , FILE*, struct fontdata*);
char identifycomment(char*);
int filtercomment(FILE*);
int rewritecomment(FILE*, FILE*, struct fontdata*) ;
char* safefgets(char*, int, FILE*);
void resetfontflags(struct fontdata*);
int embedfonts(FILE*, struct fontdata*, char*);

int emmathfnt(FILE* stmin, FILE* stmout, char* fontdir) {

	char buf[PS_BUFSIZE];
    static struct fontdata fontlist[N_FONT_FILS];
 	static char fontlistpacked = FALSE;

	if (!fontlistpacked) {
		packfontlist(fontlist);
		fontlistpacked = TRUE;
	}

    if (rewriteheader(stmin, stmout, fontlist) == BAD_EPS_HEADER) {
        return BAD_EPS_HEADER;
    }

    if (embedfonts(stmout, fontlist, fontdir) == FONT_NOT_FOUND) {
		resetfontflags(fontlist);
        return FONT_NOT_FOUND;
    }

    while (safefgets(buf, PS_BUFSIZE - 1, stmin) != NULL) {
        fprintf(stmout, "%s", buf);
    }

	resetfontflags(fontlist);

	return 0;

}

void packfontlist(struct fontdata* fontlistptr) {

	const char* fontface[] = FONT_FACE_LIST;
	int fontfacelens[N_FONT_FACS];
	int i, j;
	int famlen, namelen;
	struct fontdata* shiftptr;

	famlen = strlen(FONT_FAMILY);

	for (i = 0; i < N_FONT_FACS; ++i) {
		*(fontfacelens + i) = (int) strlen( *(fontface + i) );
	}
	
	shiftptr = fontlistptr;

	for (i = 1; i <= N_FONT_FAMS; ++i) {
		for (j = 0; j < N_FONT_FACS; ++j) {
			namelen = famlen + FAM_NUMBER + *(fontfacelens + j); 
			shiftptr->desc =
				(char *) malloc((namelen + 1) * sizeof(char));
			sprintf(shiftptr->desc, "%s%d%s", FONT_FAMILY, i, 
				*(fontface + j));
			shiftptr->inuse = 0;
			++shiftptr;
		}
	}
}

struct fontdata* findfont(struct fontdata* fontlistptr, char* fontname) {
	
	int i = 0;
	struct fontdata* retval;

	while ((i < N_FONT_FILS) && 
		strcmp((fontlistptr + i)->desc, fontname)) {
		++i;
	}

	if (i == N_FONT_FILS) {
		retval = NULL; 
	} else {
		retval = fontlistptr + i;
	}

	return retval;

}

void setfontflag(struct fontdata* fontlistptr, char fontflag) {

	fontlistptr->inuse = fontflag;
	return;

}

char isfontflagset(struct fontdata* fontlistptr) {

    return fontlistptr->inuse; 

}

void resetfontflags(struct fontdata* fontlistptr) {

	int i;

	for (i = 0; i < N_FONT_FILS; ++i) {
		fontlistptr->inuse = !FONT_IN_USE;
	}

	return;

}
		
char* fonttofilename(struct fontdata* fonthandle, char* fontdir) {

	int filenamelen;
	char *fontfile, usedefault = FALSE;

	if (fontdir == NULL) {
		usedefault = TRUE;
		fontdir = (char *) malloc( (strlen(DEF_FONT_DIR) + 1) *
			sizeof(char) );
		strcpy(fontdir, DEF_FONT_DIR);
	}
	filenamelen = strlen(fontdir) + strlen(fonthandle->desc) + 
		strlen(FONT_EXT);
	fontfile = (char *) malloc((filenamelen + 1) * sizeof(char));
	sprintf(fontfile, "%s%s%s", fontdir, fonthandle->desc, 
		FONT_EXT);
	if (usedefault) {
		free(fontdir);
	}
	return fontfile;

}

char* tokenget(char* str, int n) {

	char* token;	
	int i;

	token = strtok(str, " \t\n");

	for (i = 1; i < n; ++i) {
		token = strtok(NULL, " \t\n");
	}

	return token;
}

int rewriteheader(FILE* stmin, FILE* stmout, struct fontdata* fontlist) {

	int retval;
	long mark;
    char comid, buf[PS_BUFSIZE];

	do {
		mark = ftell(stmin);
		safefgets(buf, PS_BUFSIZE - 1, stmin);
		comid = identifycomment(buf);
		if ((comid != CODE_NO_COMMENT) && 
			(comid != CODE_OTHER) && 
			(comid != CODE_CONTINUATION)) {
			if( (comid == CODE_SUPP_RES) ||
				(comid == CODE_SUPP_FNT) ||
				(comid == CODE_NEED_FNT) ) {
				retval = filtercomment(stmin);
				if (retval == -1) {
					return BAD_EPS_HEADER;
				}
			} else {
				if (comid == CODE_NEED_RES) {
					fseek(stmin, mark, SEEK_SET);
					retval = rewritecomment(stmin, stmout, 
						fontlist);
					if (retval == -1) {
						return BAD_EPS_HEADER;
					}
				}
			}
		} else {
			if (comid != CODE_NO_COMMENT) {
				fprintf(stmout, "%s", buf);
			}
		}
	} while ((buf != NULL) && (comid != CODE_NO_COMMENT));

	if (buf == NULL) {
		return BAD_EPS_HEADER;
	}
	
	else {
		fseek(stmin, mark, SEEK_SET);
		return 0;
	}
}

/* Map resource comment to code number. */

char identifycomment(char* str) {

	char teststr[PS_BUFSIZE], *comment;
	
	strcpy(teststr, str);
	comment = tokenget(teststr, 1);

	if (*comment != '%') {
		return CODE_NO_COMMENT;
	}
	if (!strncmp(comment, DOC_NEED_RES, strlen(DOC_NEED_RES))) {
		return CODE_NEED_RES;	
	}
	if (!strncmp(comment, DOC_NEED_FNT, strlen(DOC_NEED_FNT))) {
		return  CODE_NEED_FNT;
	}
	if (!strncmp(comment, DOC_SUPP_RES, strlen(DOC_SUPP_RES))) {
		return CODE_SUPP_RES;
	}
	if (!strncmp(comment, DOC_SUPP_FNT, strlen(DOC_SUPP_FNT))) {
		return CODE_SUPP_FNT;
	}
	if (!strncmp(comment, CONTINUATION, strlen(CONTINUATION))) {
		return CODE_CONTINUATION;
	} else {
		return CODE_OTHER;
	}
	
}

/* Suppress copying of comment to output stream. */

int filtercomment(FILE* stmin) {

	long mark;
	char buf[PS_BUFSIZE];
	
	do {
		mark = ftell(stmin);
		safefgets(buf, PS_BUFSIZE - 1, stmin);
	} while ( (buf != NULL) && 
		(identifycomment(buf) == CODE_CONTINUATION) );
	
	fseek(stmin, mark, SEEK_SET);

	if (buf == NULL) {
		return -1;
	} else {
		return 0;
	}
		
}

/* 
*
* Scan for DSC header comments for needed Math fonts, record them in the
* fontdata list and then update comments to reflect embedded fonts. 
*
*/

int rewritecomment(FILE* stmin, FILE* stmout, struct fontdata* fontlist) {

	int i;
	long mark;
	char comid, comstarted = FALSE, buf[PS_BUFSIZE], *fontfamily;
 	struct fontdata* fonthandle;

	fprintf(stmout, "%s: ", DOC_NEED_RES);

	while (
		(safefgets(buf, PS_BUFSIZE - 1, stmin) != NULL) && ( 
		((comid = identifycomment(buf)) == CODE_NEED_RES) ||
		(comid == CODE_CONTINUATION) ) ) {
		mark = ftell(stmin);
		fontfamily = tokenget(buf, 3);
		fonthandle = findfont(fontlist, fontfamily);
		if (fonthandle == NULL) {
			if (!comstarted) {
				comstarted = TRUE;
				fprintf(stmout, "font %s\n", fontfamily);
			} else {
				fprintf(stmout, "%s font %s\n", CONTINUATION, fontfamily);
			}
		} else {
			setfontflag(fonthandle, FONT_IN_USE);
		}
	}

	if (buf == NULL) {
		return -1;
	}

	if (!comstarted) {
		fprintf(stmout, "\n");
	}

	fseek(stmin, mark, SEEK_SET);

	comstarted = FALSE;

	fprintf(stmout, "%s: ", DOC_SUPP_RES);
	for (i = 0; i < N_FONT_FILS; ++i) {
		if ( isfontflagset(fontlist + i) ) {
			if (!comstarted) {
				comstarted = TRUE;
				fprintf(stmout, "font %s\n", (fontlist + i)->desc);
			} else {
				fprintf(stmout, "%s font %s\n", CONTINUATION, 
					(fontlist + i)-> desc);
			}
		}
	}
	
	if (!comstarted) {
		fprintf(stmout, "\n");
	}

	return 0;

}

/* Scan through font list and copy needed Type 1 fonts to stmout. */

int embedfonts(FILE* stmout, struct fontdata* fontlist, char* fontdir) {
	
	int i;
	char buf[PS_BUFSIZE], *fontfile;
	FILE* stmfnt;

	for (i = 0; i < N_FONT_FILS; ++i) {
		if (isfontflagset(fontlist + i)) {
			fprintf(stmout, "%s%s\n", BEG_RES, (fontlist + i)->desc);
			fprintf(stmout, "%s%s\n", BEG_FNT, (fontlist + i)->desc);
			fontfile = fonttofilename(fontlist + i, fontdir);
			stmfnt = fopen(fontfile, "rb");
			if (stmfnt == NULL) {
				return FONT_NOT_FOUND;
			}
			while (safefgets(buf, PS_BUFSIZE-1, stmfnt) != NULL) {
				fprintf(stmout, "%s", buf);	
			}
			fclose(stmfnt);
			free(fontfile);
			fprintf(stmout, "\n%s%s", END_FNT, END_RES);
		}
	}

	return 0;

}

/* 
*
* Replacement for fgets() that works regardless of OS end of line
* conventions.  stream must be opened in binary mode. 
*
*/

char* safefgets(char* s, int n, FILE* stream) {

	char endofline = FALSE;
	int c, cnext, i = 0;

	while ( (endofline == FALSE) && ((c = fgetc(stream)) != EOF) && 
		(i < n) ) {
		++i; 
		if ( c == '\r' || c == '\n' ) {
			endofline = TRUE;
			if ( c == '\r' ) {
				cnext = fgetc(stream);
				if ( cnext != '\n' ) {
					ungetc(cnext, stream);
				}
			}
			c = '\n';
		}
		*s = c;
		++s;
	}

	if (i == 0) {
		return NULL;
	} else {
		*s = '\0';
		return s - i;
	}
}
